
import glob
import os

import numpy as np
import pandas as pd
import ucat
import xarray as xr

def update_vcds(data, path, max_iter=10):
    """
    Update NO2 vertical column densities (VCD) by recomputing tropospheric
    air mass factors (AMF) for Sentinel-5P/TROPOMI data. The method computes
    the mean concentration in the boundary layer from ERA-5 boundary layer
    height and TROPOMI NO2 enhancement above the background
    (i.e. NO2 - NO2_estimated_background).

    The mean concentration is used to update the TM5 a priori profiles in
    the boundary layer and to compute the air mass factor using the
    averaging kernels. Since updating the AMFs will increase the enhancement,
    iteration is used till convergence.

    The method will only update AMFs for pixels that have been detected as
    enhanced above the background by the plume detection algorithm.

    The function will recompute NO2_minus_estimated_background and the mass
    columns from the updated NO2 VCDS.
    """
    time = pd.Timestamp(data.attrs['time'])
    pattern = time.strftime(
        'S5P_OPER_AUX_CTMANA_%Y%m%dT000000_????????T000000_????????T??????.nc'
    )
    # TODO: check if taken last makes sense
    aux_filename = sorted(glob.glob(os.path.join(path, pattern)))[-1]

    # open aux filename and interpolate to remote sensing image
    aux = xr.open_dataset(aux_filename)
    aux = aux.interp(lon=data.longitude, lat=data.latitude,
                     time=pd.Timestamp(data.attrs['time']),
                     method='linear')
    aux = aux.rename(lev='layer')

    # constants
    from scipy.constants import g
    R = 287.052874  # specific gas constant for dry air
    GAMMA = -0.0065 # lapse rate

    # pressure levels in middle and at interface
    pressure_mid = aux.hyam + aux.hybm * aux.ps
    pressure_int = aux.hyai + aux.hybi * aux.ps

    # correct surface pressure for topography (Zhou et al. 2009)
    ps_eff = aux.ps * (aux.t[0] / (aux.t[0] + GAMMA * (aux.surface_altitude - data.surface_altitude))) ** (-g / R / GAMMA)

    # effective pressure levels in middle and at interface
    pressure_mid_eff = aux.hyam + aux.hybm * ps_eff
    pressure_int_eff = aux.hyai + aux.hybi * ps_eff

    # layer thickness and mid height (TODO: use virtual temperature)
    dh = R * aux.t / acceleration_due_to_gravity * np.log(pressure_int_eff[:-1].values / pressure_int_eff[1:].values)
    dh = xr.where(np.isfinite(dh), dh, 0.0)
    h = dh.cumsum(dim='layer') - 0.5 * dh

    # NO2 profile correction for topography
    no2_topo_correction = (pressure_int_eff[:-1].values - pressure_int_eff[1:].values) / (pressure_int[:-1].values - pressure_int[1:].values)
    no2 = aux.no2 * no2_topo_correction

    # Averaging kernels above tropopause to zero for tropospheric AMF
    AK = xr.where(data.averaging_kernel.layer >= aux.tropopause_layer_index,
                  data.averaging_kernel, data.averaging_kernel)

    # Convert TM5 to molar concentration (TODO: account for humidity)
    NO2_TM5_profiles = ucat.convert_points(no2, 'mol mol-1', 'mol m-3',
                                           p=pressure_mid_eff, T=aux.t)

    data = data.rename(NO2='NO2_standard',
                       air_mass_factor_troposphere='air_mass_factor_troposphere_standard')

    # Iterative update tropospheric AMFs and NO2 VCDs
    NO2_current = data['NO2_standard'].copy()
    amf_standard = data['air_mass_factor_troposphere_standard'].copy()

    for i in range(max_iter):

        # Compute NO2 concentration in boundary layer (mol / m³)
        NO2_inside_blh = ((NO2_current - data.NO2_estimated_background) / data.blh)

        NO2_profile = xr.where((h <= data.blh) & data.is_hit,
                               NO2_inside_blh, NO2_TM5_profiles)

        # Update AMFs and NO2 VCDs
        AMF_update = amf_standard * (AK * NO2_profile * dh).sum('layer') \
                     / (NO2_profile * dh).sum('layer')
        NO2_update = amf_standard * data['NO2_standard'] / AMF_update

        change = np.nanmean(
            np.abs(NO2_update.values[data.is_hit] - NO2_current.values[data.is_hit])
            / NO2_current.values[data.is_hit]
        )
        if change <= 0.01:
            break

        NO2_current, NO2_update = NO2_update, None

    else:
        raise ValueError(f'Maximum number of iterations ({max_iter}) without convergence.')

    # Add updated VCDs and AMFs to remote sensing dataset
    data['NO2'] = xr.DataArray(
        NO2_update,
        dims=data['NO2_standard'].dims,
        attrs=data['NO2_standard'].attrs
    )
    data['NO2'].attrs['comment'] = 'AMF correction applied assuming well-mixed NO2 profile for pixels with enhanced NO2 columns.'

    data['air_mass_factor_troposphere'] = xr.DataArray(
        AMF_update,
        dims=data['air_mass_factor_troposphere_standard'].dims,
        attrs=data['air_mass_factor_troposphere_standard'].attrs
    )
    data['air_mass_factor_troposphere'].attrs['comment'] = 'AMF correction applied assuming well-mixed NO2 profile for pixels with enhanced NO2 columns.'


    # Update fields from plume detection and pre-processing
    # (i.e. NO2_minus_estimated_background and *_mass)
    data['NO2_minus_estimated_background'] = data['NO2'] - data['NO2_estimated_background']

    data['NO2_mass'] = xr.DataArray(
            ucat.convert_columns(data['NO2'], 'mol m-2', 'kg m-2', molar_mass='NO2'),
            dims=data['NO2_mass'].dims, attrs=data['NO2_mass'].attrs
    )
    data['NO2_minus_estimated_background_mass'] = xr.DataArray(
            ucat.convert_columns(data['NO2_minus_estimated_background'],
                                 'mol m-2', 'kg m-2', molar_mass='NO2'),
            dims=data['NO2_minus_estimated_background_mass'].dims,
        attrs=data['NO2_minus_estimated_background_mass'].attrs
    )

    return data